/**
 * This file is part of the X10 project (http://x10-lang.org).
 *
 * This file is licensed to You under the Eclipse Public License (EPL);
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *     http://www.opensource.org/licenses/eclipse-1.0.php
 *
 * (C) Copyright IBM Corporation 2006-2014.
 */

package de.uks.plm.ftglb.incremental;

import x10.util.Team;
import x10.util.RailUtils;

import x10.compiler.Inline;

import x10.xrx.Runtime;

import de.uks.plm.utils.AssertPrinter;

/**
 * <p>The top level class of the Global Load Balancing (GLB) framework.
 * </p>
 */
public final class GLB[Queue, R]{Queue<:TaskQueue[Queue, R], R<:Arithmetic[R]}
{
    /**
     * Number of places.
     */
    private val P = Place.numPlaces();

    /**
     * Home PlaceLocalHandle of {@link Worker}
     */
    private val plh:PlaceLocalHandle[Worker[Queue, R]];

    /**
     * Workload initialization time.
     */
    var setupTime:Long;

    /**
     * Computation time.
     */
    var crunchNumberTime:Long;

    /**
     * Result collection time.
     */
    var collectResultTime:Long;

    /**
     * {@link GLBResult at root. Used as a vehicle to collect results.}
     */
    var rootGlbR: GLBResult[R] = null; // root glb result collector 

    /**
     * Min helper method.
     */
    @Inline static def min(val i:Long, val j:Long) : Long
    {
        return ((i < j) ? i : j);
    }

    /**
     * GLB Parameters. {@link GLBParameters}
     */
    var glbParams:GLBParameters;

    /**
     * Constructor
     *
     * @param init function closure that can initialize {@link TaskQueue}
     * @param glbParams GLB parameters
     * @tree true if workload is dynamically generated,
     *       false if workload can be known upfront. 
     */
    public def this(val init:()=>Queue
                    , val glbParams:GLBParameters
                    , val tree:Boolean)
    {
        AssertPrinter.println("[GLB " + here + "]: entering constructor.");
        this.glbParams = glbParams;
        this.setupTime = System.nanoTime();

        val workerInit:() => Worker[Queue, R]
            = () =>
              {
                  return (new Worker[Queue, R](init
                                               , glbParams.n
                                               , glbParams.w
                                               , glbParams.l
                                               , glbParams.z
                                               , glbParams.m
                                               , glbParams.k
                                               , glbParams.r
                                               , tree));
              };

        val group:PlaceGroup = Place.places();

        this.plh
            = PlaceLocalHandle.makeFlat[Worker[Queue, R]](group, workerInit);

        Worker.initContexts[Queue, R](plh);
        this.setupTime = System.nanoTime() - this.setupTime;
        AssertPrinter.println("[GLB " + here + "]: leaving constructor.");
    }

    /**
     * Returns Home {@link TaskQueue}
     */
    public def taskQueue() : Queue[Queue, R]
    {
        return (plh().queue);
    }

    /**
     * Run method. This method is called when users does not know the workload
     * upfront.
     *
     * @param start The method that (Root) initializes the workload that can 
     * start computation. Other places first get their workload by stealing.
     */
    public def run(val start:() => void) : Rail[R]
    {
        AssertPrinter.println("[GLB " + here + "]: starting.");
        crunchNumberTime = System.nanoTime();

        try
        {
            plh().main(start);
        }
        catch (val c:CheckedThrowable)
        {
            c.printStackTrace();
        }

        AssertPrinter.println("[GLB " + here + "]: stopping.");
        crunchNumberTime = System.nanoTime() - crunchNumberTime;
        AssertPrinter.println("[GLB " + here + "]: reducing result.");
        r:Rail[R] = collectResults();
        end(r);
        return (r);
    }

    /**
     * Run method. This method is called when users can know the workload
     * upfront and initialize the workload in {@link TaskQueue}
     */
    public def runParallel() : Rail[R]
    {
        crunchNumberTime = System.nanoTime();
        try
        {
            Worker.broadcast[Queue,R](plh);
        }
        catch (val e:CheckedThrowable)
        {
            e.printStackTrace();
        }
        crunchNumberTime = System.nanoTime() - crunchNumberTime;
        AssertPrinter.println("[GLB " + here + "]: reducing result.");
        r:Rail[R] = collectResults();
        end(r);
        return (r);
    }

    /**
     * Print various GLB-related information, including result; time spent in 
     * initialization, computation and result collection; any user specified log
     * information (per place); and GLB statistics.
     *
     * @param r result to print
     */
    private def end(val r:Rail[R]) : void
    {
        // print result
        if (0n != (glbParams.v & GLBParameters.SHOW_RESULT_FLAG))
        {
            rootGlbR.display(r);
        }
        if (0n != (glbParams.v & GLBParameters.SHOW_TIMING_FLAG))
        { // print overall timing information
            Console.OUT.println("Setup time(s):" + ((setupTime) / 1E9));

            Console.OUT.println(  "Process time(s):"
                                + ((crunchNumberTime) / 1E9));

            Console.OUT.println(  "Result reduce time(s):"
                                + (collectResultTime / 1E9));
        }

        // print log
        if (0n != (glbParams.v & GLBParameters.SHOW_TASKFRAME_LOG_FLAG))
        { 
            printLog(plh);
        }

        // collect glb statistics and print it out
        if (0n != (glbParams.v & GLBParameters.SHOW_GLB_FLAG))
        {
            collectLifelineStatus(plh);
        }
    }

    /**
     * Collect GLB statistics
     * @param st PlaceLocalHandle for {@link Worker}
     */
    private def collectLifelineStatus(val st:PlaceLocalHandle[Worker[Queue, R]])
        : void
    {
        val logs:Rail[Logger] = new Rail[Logger](P);

        val params:Boolean
            = (0n != (this.glbParams.v & GLBParameters.SHOW_GLB_FLAG));

        val places:PlaceGroup = Place.places().filterDeadPlaces();
        
        for (place:Place in places)
        {
            val id:Long = place.id();
            logs(id) = at (place)
                       {
                           return (st().logger.get(params));
                       };
        }
        // if (1024 < P)
        // {
        //     logs = new Rail[Logger]
        //         (P/32, (val i:Long) =>at (Place(i * 32))
        //                {
        //                    val h = Runtime.hereLong();
        //                    val n = min(32, P-h);
        //                    val logCollect:(val i:Long) => Logger
        //                        = (var i:Long) => 
        //                          {
        //                              val log:Logger
        //                                  = at (Place(h+i))st().logger
        //                                                        .get(params);
        //                              return (log);
        //                          };
        //                    val logs = new Rail[Logger](n, logCollect);
        //                    val log = new Logger(false);
        //                    log.collect(logs);
        //                    return (log);
        //                });
        // }
        // else
        // {
        //     logs = new Rail[Logger](P
        //                             , (i:Long) => at (Place(i))
        //                                           st().logger.get(params));
        // }
        val log = new Logger(false);
        log.collect(logs);
        log.stats();
    }
    
    /**
     * Collect results from all places and reduce them to the final result.
     * @return Final result.
     */
    protected def collectResults():Rail[R]{
        /*
         * trick taught by Dave, caputure this.plh (as a pointer) instead of
         * calling plh() directly inside the closure, which will encapsulate
         * this (i.e. the whole rail of
         */
        val tmpPlh = plh;

        collectResultTime = System.nanoTime();

        var myAlive:PlaceGroup = Place.places().filterDeadPlaces();;
        // var myAliveTeam:Team = Team(myAlive);
        var canReduce:Boolean = false;
        while (!canReduce)
        {
            AssertPrinter.println("[GLB " + here + "]: checking validity.");

            val AND:(Boolean, Boolean) => Boolean =
                (val x:Boolean, val y:Boolean) => { return (x && y); };

            val OR_RAIL:(Rail[Boolean], Rail[Boolean]) => void
                = (val r1:Rail[Boolean], val r2:Rail[Boolean]) =>
                  {
                      val size:Long = r1.size - 1;
                      for (index:Long in 0..size)
                      {
                          r1(index) |= r2(index);
                      }
                  };

            myAlive = Place.places().filterDeadPlaces();
            // myAliveTeam = Team(myAlive);
            // val aliveTeam:Team = myAliveTeam;
            val restored:Rail[Boolean] = new Rail[Boolean](P, false);

            val reducer:() => Rail[Boolean]
                = () =>
                {
                    val P:Long = tmpPlh().P;
                    var id:Long = tmpPlh().me;
                    val restoredLocal:Rail[Boolean] = new Rail[Boolean](P);

                    val lowestRestored:Long
                        = (tmpPlh().backForth.getForth() + 1) % P;

                    if (id < lowestRestored)
                    {
                        id += P;
                    }
                    for (index:Long in lowestRestored..id)
                    {
                        restoredLocal(index % P) = true;
                    }

                    AssertPrinter.println(  "[Worker " + here + ": " 
                                          + restoredLocal);

                    return (restoredLocal);
                };

                for (place:Place in myAlive)
            {
                OR_RAIL(restored, at(place) reducer());
            }
//             val restoredGlobal:GlobalRef[Rail[Boolean]]
//                 = GlobalRef[Rail[Boolean]](restored);
// 
//             val reducer:() => void
//                 = () =>
//                   {
//                       val P:Long = tmpPlh().P;
//                       var id:Long = tmpPlh().me;
// 
//                       val lowestRestored:Long
//                           = (tmpPlh().backForth.getForth() + 1) % P;
// 
//                       AssertPrinter.println(  here + ": lowest = "
//                                             + lowestRestored + ", me = " + id);
// 
//                       val restored:Rail[Boolean];
// 
//                       if (id < lowestRestored)
//                       {
//                           id += P;
//                       }
// 
//                       if (restoredGlobal.home == here)
//                       {
//                           val tmpRestoredGlobal:GlobalRef[Rail[Boolean]]
//                               = (restoredGlobal as GlobalRef[Rail[Boolean]]
//                                   {self.home == here});
// 
//                           restored = tmpRestoredGlobal.getLocalOrCopy();
//                       }
//                       else
//                       {
//                           restored = new Rail[Boolean](P);
//                       }
// 
//                       for (index:Long in lowestRestored..id)
//                       {
//                           restored(index % P) = true;
//                       }
//                       AssertPrinter.println(here + ": " + restored);
// 
//                       aliveTeam.reduce(restoredGlobal.home
//                                        , restored
//                                        , 0
//                                        , restored
//                                        , 0
//                                        , restored.size
//                                        , Team.OR);
//                   };
// 
//             AssertPrinter.println("[GLB " + here + "]: broadcasting.");
//             myAlive.broadcastFlat(reducer);
//             AssertPrinter.println(restored);

            if (RailUtils.reduce(restored, AND, true))
            {
                canReduce = true;
            }
            else
            {
                for (index:Long in 0..(restored.size -1))
                {
                    if (!(true == restored(index)))
                    {
                        val forthId:Long = (index - 1 >= 0 ? index - 1 : P - 1);
                        val forth:Place = Place(forthId);

                        AssertPrinter.println(  "[GLB " + here + "]: "
                                              + Place(index) + " died, "
                                              + " contacting " + forth + ".");

                        val restartTime:Long = System.nanoTime();
                        finish
                        {
                            at (forth) async
                            {
                                tmpPlh().isDead(index);
                            }
                        }
                        crunchNumberTime += System.nanoTime() - restartTime;
                        AssertPrinter.println(  "[GLB " + here + "]: "
                                              + "restore of " + Place(index)
                                              + " ended successfully.");
                    }
                }
            }
        }

        this.rootGlbR = plh().queue.getResult();
        val resultGlobal = GlobalRef[GLBResult[R]](rootGlbR);
        val tmpRail:Rail[R] = rootGlbR.submitResult();
        val size:Long = tmpRail.size - 1;

        for (place:Place in myAlive)
        {
            if (place != resultGlobal.home)
            {
                val localResult:Rail[R]
                    = at (place) tmpPlh().queue.getResult().getResult();

                for(index:Long in 0..size)
                {
                    tmpRail(index) += localResult(index);
                }
            }
        }

//        val alive:PlaceGroup = myAlive;
//        val aliveTeam:Team = myAliveTeam;
//         val reducer:() => void
//             = () =>
//               {
//                   val glbR:GLBResult[R];
//                   if (here == resultGlobal.home)
//                   {
//                       val tmpResultGlobal:GlobalRef[GLBResult[R]]
//                           = (resultGlobal as GlobalRef[GLBResult[R]]
//                               { self.home == here});
//                       glbR =  tmpResultGlobal.getLocalOrCopy();
//                   }
//                   else
//                   {
//                       glbR = tmpPlh().queue.getResult();
//                   }
// 
//                   aliveTeam.allreduce(glbR.submitResult()
//                                       // Offset into the source buffer.
//                                       , 0
//                                       // Destination buffer.
//                                       , glbR.submitResult()
//                                       // Offset into the destination
//                                       // buffer.
//                                       , 0
//                                       // Number of elements.
//                                       , glbR.submitResult().size
//                                       // Operation to be performed.
//                                       , glbR.getReduceOperator());
//               };
//         alive.broadcastFlat(reducer);
        collectResultTime = System.nanoTime() - collectResultTime;

        return (tmpRail);
    }

    /**
     * Print logging information on each place if user is interested in
     * collecting per place information, i.e., statistics instrumented.
     * @param st PLH for {@link Worker}
     */
    private def printLog(val st:PlaceLocalHandle[Worker[Queue, R]]) : void
    {
        val places:PlaceGroup = Place.places().filterDeadPlaces();
        for(place:Place in places)
        {
            at(place)
            {
                st().queue.printLog();
            }
        }
    }
}
